"""
Tests of neo.io.nixio_fr
"""

import numpy as np
import unittest
from quantities import s
from neo.io.nixio_fr import NixIO as NixIOfr
import quantities as pq
from neo.io.nixio import NixIO
from neo.test.iotest.common_io_test import BaseTestIO
from neo.core import Block, Segment, AnalogSignal, SpikeTrain, Event

try:
    import nixio as nix

    HAVE_NIX = True
except ImportError:
    HAVE_NIX = False
import os


@unittest.skipUnless(HAVE_NIX, "Requires NIX")
class TestNixfr(
    BaseTestIO,
    unittest.TestCase,
):
    ioclass = NixIOfr

    entities_to_download = ["nix/nixio_fr.nix"]
    entities_to_test = ["nix/nixio_fr.nix"]

    def setUp(self):
        super().setUp()
        self.testfilename = self.get_local_path("nix/nixio_fr.nix")
        self.reader_fr = NixIOfr(filename=self.testfilename)
        self.reader_norm = NixIO(filename=self.testfilename, mode="ro")
        self.blk = self.reader_fr.read_block(block_index=1, load_waveforms=True)
        # read block with NixIOfr
        self.blk1 = self.reader_norm.read_block(index=1)  # read same block with NixIO

    def tearDown(self):
        self.reader_fr.file.close()
        self.reader_norm.close()

    def test_check_same_neo_structure(self):
        self.assertEqual(len(self.blk.segments), len(self.blk1.segments))
        for seg1, seg2 in zip(self.blk.segments, self.blk1.segments):
            self.assertEqual(len(seg1.analogsignals), len(seg2.analogsignals))
            self.assertEqual(len(seg1.spiketrains), len(seg2.spiketrains))
            self.assertEqual(len(seg1.events), len(seg2.events))
            self.assertEqual(len(seg1.epochs), len(seg2.epochs))

    def test_check_same_data_content(self):
        for seg1, seg2 in zip(self.blk.segments, self.blk1.segments):
            for asig1, asig2 in zip(seg1.analogsignals, seg2.analogsignals):
                np.testing.assert_almost_equal(asig1.magnitude, asig2.magnitude)
                # not completely equal
            for st1, st2 in zip(seg1.spiketrains, seg2.spiketrains):
                np.testing.assert_array_equal(st1.magnitude, st2.times)
                for wf1, wf2 in zip(st1.waveforms, st2.waveforms):
                    np.testing.assert_array_equal(wf1.shape, wf2.shape)
                    np.testing.assert_almost_equal(wf1.magnitude, wf2.magnitude)
            for ev1, ev2 in zip(seg1.events, seg2.events):
                np.testing.assert_almost_equal(ev1.times, ev2.times)
                assert np.all(ev1.labels == ev2.labels)
            for ep1, ep2 in zip(seg1.epochs, seg2.epochs):
                assert len(ep1.durations) == len(ep2.times)
                np.testing.assert_almost_equal(ep1.times, ep2.times)
                np.testing.assert_array_equal(ep1.durations, ep2.durations)
                np.testing.assert_array_equal(ep1.labels, ep2.labels)

    def test_analog_signal(self):
        seg1 = self.blk.segments[0]
        an_sig1 = seg1.analogsignals[0]
        assert len(an_sig1) == 30
        an_sig2 = seg1.analogsignals[1]
        assert an_sig2.shape == (50, 3)

    def test_spike_train(self):
        st1 = self.blk.segments[0].spiketrains[0]
        assert np.all(st1.times == np.cumsum(np.arange(0, 1, 0.1)).tolist() * pq.s + 10 * pq.s)

    def test_event(self):
        seg1 = self.blk.segments[0]
        event1 = seg1.events[0]
        raw_time = 10 + np.cumsum(np.array([0, 1, 2, 3, 4]))
        assert np.all(event1.times == np.array(raw_time * pq.s / 1000))
        assert np.all(event1.labels == np.array(["A", "B", "C", "D", "E"], dtype="U"))
        assert len(seg1.events) == 1

    def test_epoch(self):
        seg1 = self.blk.segments[1]
        seg2 = self.blk1.segments[1]
        epoch1 = seg1.epochs[0]
        epoch2 = seg2.epochs[0]
        assert len(epoch1.durations) == len(epoch1.times)
        assert np.all(epoch1.durations == epoch2.durations)
        assert np.all(epoch1.labels == epoch2.labels)

    def test_annotations(self):
        self.testfilename = self.get_local_path("nix/nixio_fr_ann.nix")
        with NixIO(filename=self.testfilename, mode="ow") as io:
            annotations = {"my_custom_annotation": "hello block"}
            bl = Block(**annotations)
            annotations = {"something": "hello hello000"}
            seg = Segment(**annotations)
            an = AnalogSignal([[1, 2, 3], [4, 5, 6]], units="V", sampling_rate=1 * pq.Hz)
            an.annotate(ansigrandom="hello chars")
            an.array_annotate(custom_id=[1, 2, 3])
            sp = SpikeTrain([3, 4, 5] * s, t_stop=10.0)
            sp.annotations["railway"] = "hello train"
            ev = Event(np.arange(0, 30, 10) * pq.Hz, labels=np.array(["trig0", "trig1", "trig2"], dtype="U"))
            ev.annotations["venue"] = "hello event"
            ev2 = Event(np.arange(0, 30, 10) * pq.Hz, labels=np.array(["trig0", "trig1", "trig2"], dtype="U"))
            ev2.annotations["evven"] = "hello ev"
            seg.spiketrains.append(sp)
            seg.events.append(ev)
            seg.events.append(ev2)
            seg.analogsignals.append(an)
            bl.segments.append(seg)
            io.write_block(bl)
            io.close()

        with NixIOfr(filename=self.testfilename) as frio:
            frbl = frio.read_block()
            assert "my_custom_annotation" in frbl.annotations
            assert "something" in frbl.segments[0].annotations
            assert "ansigrandom" in frbl.segments[0].analogsignals[0].annotations
            assert "railway" in frbl.segments[0].spiketrains[0].annotations
            assert "venue" in frbl.segments[0].events[0].annotations
            assert "evven" in frbl.segments[0].events[1].annotations
            assert "custom_id" in frbl.segments[0].analogsignals[0].array_annotations
            for anasig in frbl.segments[0].analogsignals:
                for value in anasig.array_annotations.values():
                    assert anasig.shape[-1] == len(value)
        os.remove(self.testfilename)


if __name__ == "__main__":
    unittest.main()



"""
This module defines :class:`Epoch`, an array of epochs.

:class:`Epoch` derives from :class:`BaseNeo`, from
:module:`neo.core.baseneo`.
"""

from copy import deepcopy
from numbers import Number

import numpy as np
import quantities as pq

from neo.core.baseneo import BaseNeo, merge_annotations
from neo.core.dataobject import DataObject, ArrayDict


def _new_epoch(
    cls,
    times=None,
    durations=None,
    labels=None,
    units=None,
    name=None,
    description=None,
    file_origin=None,
    array_annotations=None,
    annotations=None,
    segment=None,
):
    """
    A function to map epoch.__new__ to function that
    does not do the unit checking. This is needed for pickle to work.
    """
    e = Epoch(
        times=times,
        durations=durations,
        labels=labels,
        units=units,
        name=name,
        file_origin=file_origin,
        description=description,
        array_annotations=array_annotations,
        **annotations,
    )
    e.segment = segment
    return e


class Epoch(DataObject):
    """
    Array of epochs.

    Parameters
    ----------
    times: quantity array 1D | numpy array 1D | list | None, default: None
        The start times of each time period.
        If None, generates an empty array
    durations: quantity array 1D | numpy array 1D | list |  quantity scalar | float | None, default: None
        The length(s) of each time period.
        If a scalar/float, the same value is used for all time periods.
        If None, generates an empty array
    labels: numpy.array 1D dtype='U' | list | None, default: None
        Names or labels for the time periods.
        If None, creates an empty array
    units: quantity units | str | None, default: None
        The units for the time
        Required if the times is a list or NumPy, not required if it is a :class:`Quantity`
    name: (str) A label for the dataset,
    description: (str) Text description,
    file_origin: (str) Filesystem path or URL of the original data file.

    *Usage*::

        >>> from neo.core import Epoch
        >>> from quantities import s, ms
        >>> import numpy as np
        >>>
        >>> epc = Epoch(times=np.arange(0, 30, 10)*s,
        ...             durations=[10, 5, 7]*ms,
        ...             labels=np.array(['btn0', 'btn1', 'btn2'], dtype='U'))
        >>>
        >>> epc.times
        array([  0.,  10.,  20.]) * s
        >>> epc.durations
        array([ 10.,   5.,   7.]) * ms
        >>> epc.labels
        array(['btn0', 'btn1', 'btn2'],
              dtype='<U4')



    *Recommended attributes/properties*:


    *Optional attributes/properties*:
        :array_annotations: (dict) Dict mapping strings to numpy arrays containing annotations \
                                   for all data points

    Note: Any other additional arguments are assumed to be user-specific
    metadata and stored in :attr:`annotations`,

    """

    _parent_objects = ("Segment",)
    _parent_attrs = ("segment",)
    _quantity_attr = "times"
    _necessary_attrs = (
        ("times", pq.Quantity, 1),
        ("durations", pq.Quantity, 1),
        ("labels", np.ndarray, 1, np.dtype("U")),
    )

    def __new__(
        cls,
        times=None,
        durations=None,
        labels=None,
        units=None,
        name=None,
        description=None,
        file_origin=None,
        array_annotations=None,
        **annotations,
    ):
        if times is None:
            times = np.array([]) * pq.s
        elif isinstance(times, (list, tuple)):
            times = np.array(times)
        if len(times.shape) > 1:
            raise ValueError("Times array has more than 1 dimension")
        if isinstance(durations, (list, tuple)):
            durations = np.array(durations)
        elif durations is None:
            durations = np.array([]) * pq.s
        elif isinstance(durations, Number):
            durations = durations * np.ones(times.shape)
        elif durations.size != times.size:
            if durations.size == 1:
                durations = durations * np.ones_like(times.magnitude)
            else:
                raise ValueError("Durations array has different length to times")
        if labels is None:
            labels = np.array([], dtype="U")
        else:
            labels = np.array(labels)
            if labels.size != times.size and labels.size:
                raise ValueError("Labels array has different length to times")
        if units is None:
            # No keyword units, so get from `times`
            try:
                units = times.units
                dim = units.dimensionality
            except AttributeError:
                raise ValueError("you must specify units")
        else:
            if hasattr(units, "dimensionality"):
                dim = units.dimensionality
            else:
                dim = pq.quantity.validate_dimensionality(units)
        if not hasattr(durations, "dimensionality"):
            durations = pq.Quantity(durations, dim)
        # check to make sure the units are time
        # this approach is much faster than comparing the
        # reference dimensionality
        if len(dim) != 1 or list(dim.values())[0] != 1 or not isinstance(list(dim.keys())[0], pq.UnitTime):
            ValueError(f"Unit {units} has dimensions {dim.simplified}, not [time]")

        obj = pq.Quantity.__new__(cls, times, units=dim)
        obj._labels = labels
        obj._durations = durations
        obj.segment = None
        return obj

    def __init__(
        self,
        times=None,
        durations=None,
        labels=None,
        units=None,
        name=None,
        description=None,
        file_origin=None,
        array_annotations=None,
        **annotations,
    ):
        """
        Initialize a new :class:`Epoch` instance.
        """
        DataObject.__init__(
            self,
            name=name,
            file_origin=file_origin,
            description=description,
            array_annotations=array_annotations,
            **annotations,
        )

    def __reduce__(self):
        """
        Map the __new__ function onto _new_epoch, so that pickle
        works
        """
        return _new_epoch, (
            self.__class__,
            self.times,
            self.durations,
            self.labels,
            self.units,
            self.name,
            self.file_origin,
            self.description,
            self.array_annotations,
            self.annotations,
            self.segment,
        )

    def __array_finalize__(self, obj):
        super().__array_finalize__(obj)
        self._durations = getattr(obj, "durations", None)
        self._labels = getattr(obj, "labels", None)
        self.annotations = getattr(obj, "annotations", None)
        self.name = getattr(obj, "name", None)
        self.file_origin = getattr(obj, "file_origin", None)
        self.description = getattr(obj, "description", None)
        self.segment = getattr(obj, "segment", None)
        # Add empty array annotations, because they cannot always be copied,
        # but do not overwrite existing ones from slicing etc.
        # This ensures the attribute exists
        if not hasattr(self, "array_annotations"):
            self.array_annotations = ArrayDict(self._get_arr_ann_length())

    def __repr__(self):
        """
        Returns a string representing the :class:`Epoch`.
        """

        objs = [
            f"{label}@{str(time)} for {str(dur)}" for label, time, dur in zip(self.labels, self.times, self.durations)
        ]
        return "<Epoch: %s>" % ", ".join(objs)

    def _repr_pretty_(self, pp, cycle):
        labels = ""
        if self._labels is not None:
            labels = " with labels"
        pp.text(
            f"{self.__class__.__name__} containing {self.size} epochs{labels}; "
            f"time units {self.units.dimensionality.string}; datatype {self.dtype} "
        )
        if self._has_repr_pretty_attrs_():
            pp.breakable()
            self._repr_pretty_attrs_(pp, cycle)

    def rescale(self, units, dtype=None):
        """
        Return a copy of the :class:`Epoch` converted to the specified units
        The `dtype` argument exists only for backward compatibility within quantities, see
        https://github.com/python-quantities/python-quantities/pull/204
        :return: Copy of self with specified units
        """
        # Use simpler functionality, if nothing will be changed
        dim = pq.quantity.validate_dimensionality(units)
        if self.dimensionality == dim:
            return self.copy()

        # Rescale the object into a new object
        obj = self.duplicate_with_new_data(
            times=self.view(pq.Quantity).rescale(dim),
            durations=self.durations.rescale(dim),
            labels=self.labels,
            units=units,
        )

        # Expected behavior is deepcopy, so deepcopying array_annotations
        obj.array_annotations = deepcopy(self.array_annotations)
        obj.segment = self.segment
        return obj

    def __getitem__(self, i):
        """
        Get the item or slice :attr:`i`.
        """
        obj = super().__getitem__(i)
        obj._durations = self.durations[i]
        if self._labels is not None and self._labels.size > 0:
            obj._labels = self.labels[i]
        else:
            obj._labels = self.labels
        try:
            # Array annotations need to be sliced accordingly
            obj.array_annotate(**deepcopy(self.array_annotations_at_index(i)))
            obj._copy_data_complement(self)
        except AttributeError:  # If Quantity was returned, not Epoch
            obj.times = obj
            obj.durations = obj._durations
            obj.labels = obj._labels
        return obj

    def __getslice__(self, i, j):
        """
        Get a slice from :attr:`i` to :attr:`j`.attr[0]

        Doesn't get called in Python 3, :meth:`__getitem__` is called instead
        """
        return self.__getitem__(slice(i, j))

    @property
    def times(self):
        return pq.Quantity(self)

    def merge(self, other):
        """
        Merge the another :class:`Epoch` into this one.

        The :class:`Epoch` objects are concatenated horizontally
        (column-wise), :func:`np.hstack`).

        If the attributes of the two :class:`Epoch` are not
        compatible, and Exception is raised.
        """
        othertimes = other.times.rescale(self.times.units)
        otherdurations = other.durations.rescale(self.durations.units)
        times = np.hstack([self.times, othertimes]) * self.times.units
        durations = np.hstack([self.durations, otherdurations]) * self.durations.units
        labels = np.hstack([self.labels, other.labels])
        kwargs = {}
        for name in ("name", "description", "file_origin"):
            attr_self = getattr(self, name)
            attr_other = getattr(other, name)
            if attr_self == attr_other:
                kwargs[name] = attr_self
            else:
                kwargs[name] = f"merge({attr_self}, {attr_other})"

        merged_annotations = merge_annotations(self.annotations, other.annotations)
        kwargs.update(merged_annotations)

        kwargs["array_annotations"] = self._merge_array_annotations(other)

        return Epoch(times=times, durations=durations, labels=labels, **kwargs)

    def _copy_data_complement(self, other):
        """
        Copy the metadata from another :class:`Epoch`.
        Note: Array annotations can not be copied here because length of data can change
        """
        # Note: Array annotations cannot be copied because length of data could be changed
        # here which would cause inconsistencies. This is instead done locally.
        for attr in ("name", "file_origin", "description"):
            setattr(self, attr, deepcopy(getattr(other, attr, None)))
        self._copy_annotations(other)

    def _copy_annotations(self, other):
        self.annotations = deepcopy(other.annotations)

    def duplicate_with_new_data(self, times, durations, labels, units=None):
        """
        Create a new :class:`Epoch` with the same metadata
        but different data (times, durations)

        Note: Array annotations can not be copied here because length of data can change
        """

        if units is None:
            units = self.units
        else:
            units = pq.quantity.validate_dimensionality(units)

        new = self.__class__(times=times, durations=durations, labels=labels, units=units)
        new._copy_data_complement(self)
        new._labels = labels
        new._durations = durations
        # Note: Array annotations can not be copied here because length of data can change
        return new

    def time_slice(self, t_start, t_stop):
        """
        Creates a new :class:`Epoch` corresponding to the time slice of
        the original :class:`Epoch` between (and including) times
        :attr:`t_start` and :attr:`t_stop`. Either parameter can also be None
        to use infinite endpoints for the time interval.
        """
        _t_start = t_start
        _t_stop = t_stop
        if t_start is None:
            _t_start = -np.inf
        if t_stop is None:
            _t_stop = np.inf

        indices = (self >= _t_start) & (self <= _t_stop)

        # Time slicing should create a deep copy of the object
        new_epc = deepcopy(self[indices])

        return new_epc

    def time_shift(self, t_shift):
        """
        Shifts an :class:`Epoch` by an amount of time.

        Parameters
        ----------
        t_shift: Quantity (time)
            Amount of time by which to shift the :class:`Epoch`.

        Returns
        -------
        epoch: :class:`Epoch`
            New instance of an :class:`Epoch` object starting at t_shift later than the
            original :class:`Epoch` (the original :class:`Epoch` is not modified).
        """
        new_epc = self.duplicate_with_new_data(times=self.times + t_shift, durations=self.durations, labels=self.labels)

        # Here we can safely copy the array annotations since we know that
        # the length of the Epoch does not change.
        new_epc.array_annotate(**self.array_annotations)

        return new_epc

    def set_labels(self, labels):
        if self.labels is not None and self.labels.size > 0 and len(labels) != self.size:
            raise ValueError(f"Labels array has different length to times " f"({len(labels)} != {self.size})")
        self._labels = np.array(labels)

    def get_labels(self):
        return self._labels

    labels = property(get_labels, set_labels)

    def set_durations(self, durations):
        if self.durations is not None and self.durations.size > 0 and len(durations) != self.size:
            raise ValueError("Durations array has different length to times " f"({len(durations)} != {self.size})")
        self._durations = durations

    def get_durations(self):
        return self._durations

    durations = property(get_durations, set_durations)



"""
Tests of neo.rawio.axographrawio
"""

import unittest

from neo.rawio.axographrawio import AxographRawIO
from neo.test.rawiotest.common_rawio_test import BaseTestRawIO


class TestAxographRawIO(BaseTestRawIO, unittest.TestCase):
    rawioclass = AxographRawIO
    entities_to_test = [
        "axograph/AxoGraph_Graph_File",  # version 1 file, provided with AxoGraph
        "axograph/AxoGraph_Digitized_File",  # version 2 file, provided with AxoGraph
        "axograph/AxoGraph_X_File.axgx",  # version 5 file, provided with AxoGraph
        "axograph/File_axograph.axgd",  # version 6 file
        "axograph/episodic.axgd",
        "axograph/events_and_epochs.axgx",
        "axograph/written-by-axographio-with-linearsequence.axgx",
        "axograph/written-by-axographio-without-linearsequence.axgx",
        "axograph/corrupt-comment.axgx",
    ]
    entities_to_download = ["axograph"]


if __name__ == "__main__":
    unittest.main()



from neo.io.basefromrawio import BaseFromRaw
from neo.rawio.micromedrawio import MicromedRawIO
from neo.core import Segment, AnalogSignal, Epoch, Event


class MicromedIO(MicromedRawIO, BaseFromRaw):
    """Class for reading/writing data from Micromed files (.trc)."""

    _prefered_signal_group_mode = "group-by-same-units"

    def __init__(self, filename):
        MicromedRawIO.__init__(self, filename=filename)
        BaseFromRaw.__init__(self, filename)



"""
Tests of neo.io.klustakwikio
"""

from pathlib import Path
import shutil

import unittest

import numpy as np
import quantities as pq

import neo
from neo.test.iotest.common_io_test import BaseTestIO
from neo.test.tools import assert_arrays_almost_equal
from neo.io.klustakwikio import KlustaKwikIO


class KlustaKwikTests(BaseTestIO, unittest.TestCase):
    ioclass = KlustaKwikIO
    entities_to_download = ["klustakwik"]
    entities_to_test = [
        "klustakwik/test2/base",
        "klustakwik/test2/base2",
    ]

    @classmethod
    def setUpClass(cls, *args, **kwargs):
        super(KlustaKwikTests, cls).setUpClass(*args, **kwargs)
        dirname = Path(cls.get_local_path("klustakwik"))

        cls.session1 = dirname / "test1"
        cls.session2 = dirname / "test2"
        cls.session3 = dirname / "test3"
        cls.tmp_session = dirname / "tmp_session"
        cls.tmp_session.mkdir(exist_ok=True)

    def tearDown(self) -> None:
        shutil.rmtree(self.tmp_session, ignore_errors=True)
        self.tmp_session.mkdir(exist_ok=True)

    def test_load_by_basename(self):
        """Tests that files can be loaded by basename"""
        kio = KlustaKwikIO(dirname=self.session1 / "basename")
        fetfiles = kio._fp.read_filenames("fet")

        self.assertEqual(len(fetfiles), 2)
        self.assertEqual(Path(fetfiles[0]).absolute(), self.session1 / "basename.fet.0")
        self.assertEqual(Path(fetfiles[1]).absolute(), self.session1 / "basename.fet.1")

    # def test_load_without_basename(self):
    #     """Tests that files are loaded even without basename"""
    #     pass
    #
    #     # this test is in flux, should probably have it default to
    #     # basename = os.path.split(dirname)[1] when dirname is a directory)
    #     kio = KlustaKwikIO(dirname=self.session1)
    #     fetfiles = kio._fp.read_filenames('fet')
    #
    #     # It will just choose one of the two basenames, depending on which
    #     # is first, so just assert that it did something without error.
    #     self.assertNotEqual(len(fetfiles), 0)

    def test_load_by_basename_2(self):
        """Tests that files can be loaded by basename2"""
        kio = KlustaKwikIO(dirname=self.session1 / "basename2")
        clufiles = kio._fp.read_filenames("clu")

        self.assertEqual(len(clufiles), 1)
        self.assertEqual(Path(clufiles[1]).absolute(), self.session1 / "basename2.clu.1")

    def test_read_data(self):
        """Tests that data and metadata are read correctly"""
        kio = KlustaKwikIO(dirname=self.session2 / "base", sampling_rate=1000.0)
        block = kio.read()[0]
        seg = block.segments[0]
        self.assertEqual(len(seg.spiketrains), 4)

        for st in seg.spiketrains:
            self.assertEqual(st.units, np.array(1.0) * pq.s)
            self.assertEqual(st.t_start, 0.0)

        self.assertEqual(seg.spiketrains[0].name, "unit 1 from group 0")
        self.assertEqual(seg.spiketrains[0].annotations["cluster"], 1)
        self.assertEqual(seg.spiketrains[0].annotations["group"], 0)
        self.assertTrue(np.all(seg.spiketrains[0].times == np.array([0.100, 0.200])))

        self.assertEqual(seg.spiketrains[1].name, "unit 2 from group 0")
        self.assertEqual(seg.spiketrains[1].annotations["cluster"], 2)
        self.assertEqual(seg.spiketrains[1].annotations["group"], 0)
        self.assertEqual(seg.spiketrains[1].t_start, 0.0)
        self.assertTrue(np.all(seg.spiketrains[1].times == np.array([0.305])))

        self.assertEqual(seg.spiketrains[2].name, "unit -1 from group 1")
        self.assertEqual(seg.spiketrains[2].annotations["cluster"], -1)
        self.assertEqual(seg.spiketrains[2].annotations["group"], 1)
        self.assertEqual(seg.spiketrains[2].t_start, 0.0)
        self.assertTrue(np.all(seg.spiketrains[2].times == np.array([0.253])))

        self.assertEqual(seg.spiketrains[3].name, "unit 2 from group 1")
        self.assertEqual(seg.spiketrains[3].annotations["cluster"], 2)
        self.assertEqual(seg.spiketrains[3].annotations["group"], 1)
        self.assertEqual(seg.spiketrains[3].t_start, 0.0)
        self.assertTrue(np.all(seg.spiketrains[3].times == np.array([0.050, 0.152])))

    def test_default_cluster_id_0(self):
        """Checks that cluster id autosets to 0 without clu file"""
        kio = KlustaKwikIO(dirname=self.session2 / "base2", sampling_rate=1000.0)
        block = kio.read()[0]
        seg = block.segments[0]
        self.assertEqual(len(seg.spiketrains), 1)
        self.assertEqual(seg.spiketrains[0].name, "unit 0 from group 5")
        self.assertEqual(seg.spiketrains[0].annotations["cluster"], 0)
        self.assertEqual(seg.spiketrains[0].annotations["group"], 5)
        self.assertEqual(seg.spiketrains[0].t_start, 0.0)
        self.assertTrue(np.all(seg.spiketrains[0].times == np.array([0.026, 0.122, 0.228])))

    def test_write_clu_and_fet(self):
        """Create clu and fet files based on spiketrains in a block.

        Checks that
            Files are created
            Converted to samples correctly
            Missing sampling rate are taken from IO reader default
            Spiketrains without cluster info are assigned to cluster 0
            Spiketrains across segments are concatenated
        """
        block = neo.Block()
        segment = neo.Segment()
        segment2 = neo.Segment()
        block.segments.append(segment)
        block.segments.append(segment2)

        # Fake spiketrain 1, will be sorted
        st1 = neo.SpikeTrain(times=[0.002, 0.004, 0.006], units="s", t_stop=1.0)
        st1.annotations["cluster"] = 0
        st1.annotations["group"] = 0
        segment.spiketrains.append(st1)

        # Fake spiketrain 1B, on another segment. No group specified,
        # default is 0.
        st1B = neo.SpikeTrain(times=[0.106], units="s", t_stop=1.0)
        st1B.annotations["cluster"] = 0
        segment2.spiketrains.append(st1B)

        # Fake spiketrain 2 on same group, no sampling rate specified
        st2 = neo.SpikeTrain(times=[0.001, 0.003, 0.011], units="s", t_stop=1.0)
        st2.annotations["cluster"] = 1
        st2.annotations["group"] = 0
        segment.spiketrains.append(st2)

        # Fake spiketrain 3 on new group, with different sampling rate
        st3 = neo.SpikeTrain(times=[0.05, 0.09, 0.10], units="s", t_stop=1.0)
        st3.annotations["cluster"] = -1
        st3.annotations["group"] = 1
        segment.spiketrains.append(st3)

        # Fake spiketrain 4 on new group, without cluster info
        st4 = neo.SpikeTrain(times=[0.005, 0.009], units="s", t_stop=1.0)
        st4.annotations["group"] = 2
        segment.spiketrains.append(st4)

        # Create writer with default sampling rate
        kio = KlustaKwikIO(dirname=self.tmp_session / "base1", sampling_rate=1000.0)
        kio.write_block(block)

        # Check files were created
        for fn in [".fet.0", ".fet.1", ".clu.0", ".clu.1"]:
            self.assertTrue((self.tmp_session / ("base1" + fn)).exists())

        # Check files contain correct content
        # Spike times on group 0
        with open(self.tmp_session / "base1.fet.0", mode="r") as f:
            data = f.readlines()
        data = [int(d) for d in data]
        self.assertEqual(data, [0, 2, 4, 6, 1, 3, 11, 106])

        # Clusters on group 0
        with open(self.tmp_session / "base1.clu.0", mode="r") as f:
            data = f.readlines()
        data = [int(d) for d in data]
        self.assertEqual(data, [2, 0, 0, 0, 1, 1, 1, 0])

        # Spike times on group 1
        with open(self.tmp_session / "base1.fet.1", mode="r") as f:
            data = f.readlines()
        data = [int(d) for d in data]
        self.assertEqual(data, [0, 50, 90, 100])

        # Clusters on group 1
        with open(self.tmp_session / "base1.clu.1") as f:
            data = f.readlines()
        data = [int(d) for d in data]
        self.assertEqual(data, [1, -1, -1, -1])

        # Spike times on group 2
        with open(self.tmp_session / "base1.fet.2") as f:
            data = f.readlines()
            data = [int(d) for d in data]
        self.assertEqual(data, [0, 5, 9])

        # Clusters on group 2
        with open(self.tmp_session / "base1.clu.2") as f:
            data = f.readlines()
        data = [int(d) for d in data]
        self.assertEqual(data, [1, 0, 0])

    def test_write_clu_and_fet_1(self):
        """Create clu and fet files based on spiketrains in a block.

        Checks that
            Files are created
            Converted to samples correctly
            Missing sampling rate are taken from IO reader default
            Spiketrains without cluster info are assigned to cluster 0
            Spiketrains across segments are concatenated
        """
        block = neo.Block()
        segment = neo.Segment()
        segment2 = neo.Segment()
        block.segments.append(segment)
        block.segments.append(segment2)

        # Fake spiketrain 1
        st1 = neo.SpikeTrain(times=[0.002, 0.004, 0.006], units="s", t_stop=1.0)
        st1.annotations["cluster"] = 0
        st1.annotations["group"] = 0
        wff = np.array([[11.3, 0.2], [-0.3, 12.3], [3.0, -2.5]])
        st1.annotations["waveform_features"] = wff
        segment.spiketrains.append(st1)

        # Create writer
        kio = KlustaKwikIO(dirname=self.tmp_session / "base2", sampling_rate=1000.0)
        kio.write_block(block)

        # Check files were created
        for fn in [".fet.0", ".clu.0"]:
            self.assertTrue(self.tmp_session / ("base2" + fn))

        # Check files contain correct content
        with open(self.tmp_session / "base2.fet.0") as fi:
            # first line is nbFeatures
            self.assertEqual(fi.readline(), "2\n")

            # Now check waveforms and times are same
            data = fi.readlines()
            new_wff = []
            new_times = []
            for line in data:
                line_split = line.split()
                new_wff.append([float(val) for val in line_split[:-1]])
                new_times.append(int(line_split[-1]))
            self.assertEqual(new_times, [2, 4, 6])
            assert_arrays_almost_equal(wff, np.array(new_wff), 0.00001)

        # Clusters on group 0
        with open(self.tmp_session / "base2.clu.0") as fi:
            data = fi.readlines()
            data = [int(d) for d in data]
            self.assertEqual(data, [1, 0, 0, 0])

            # Now read the features and test same
            block = kio.read_block()
            train = block.segments[0].spiketrains[0]
            assert_arrays_almost_equal(wff, train.annotations["waveform_features"], 0.00001)


if __name__ == "__main__":
    unittest.main()



"""
Tests of the neo.core.block.Block class
"""

from datetime import datetime
from copy import deepcopy
from neo.core.view import ChannelView

import unittest

import numpy as np

try:
    from IPython.lib.pretty import pretty
except ImportError as err:
    HAVE_IPYTHON = False
else:
    HAVE_IPYTHON = True

from neo.core.block import Block
from neo.core.segment import Segment
from neo.core.container import filterdata
from neo.core import SpikeTrain, AnalogSignal, Event
from neo.test.tools import assert_neo_object_is_compliant, assert_same_sub_schema
from neo.test.generate_datasets import random_block, simple_block, random_signal


N_EXAMPLES = 5


class TestBlock(unittest.TestCase):

    def setUp(self):
        self.blocks = [random_block() for i in range(N_EXAMPLES)]

    def test_block_init(self):
        blk = Block(name="a block")
        assert_neo_object_is_compliant(blk)
        self.assertEqual(blk.name, "a block")
        self.assertEqual(blk.file_origin, None)

    def test__merge(self):
        blk1 = self.blocks[0]
        blk2 = self.blocks[1]

        orig_blk1 = deepcopy(blk1)
        blk1.merge(blk2)

        assert_same_sub_schema(orig_blk1.segments + blk2.segments, blk1.segments)

    def test__size(self):
        for block in self.blocks:
            targ = {
                "segments": len(block.segments),
                "groups": len(block.groups),  # only counts the top-level groups?
            }
            self.assertEqual(block.size, targ)

    def test__filter_none(self):
        for block in self.blocks:
            targ = []
            # collecting all data objects in target block
            for seg in block.segments:
                targ.extend(seg.analogsignals)
                targ.extend(seg.epochs)
                targ.extend(seg.events)
                targ.extend(seg.irregularlysampledsignals)
                targ.extend(seg.spiketrains)
                targ.extend(seg.imagesequences)
            chv_names = set([])
            roi_names = set([])
            for grp in block.groups:
                for grp1 in grp.walk():
                    for chv in grp1.channelviews:
                        if chv.name not in chv_names:
                            targ.append(chv)
                            chv_names.add(chv.name)
                    for roi in grp1.regionsofinterest:
                        if roi.name not in roi_names:
                            targ.append(roi)
                            roi_names.add(roi.name)

            res1 = block.filter()
            res2 = block.filter({})
            res3 = block.filter([])
            res4 = block.filter([{}])
            res5 = block.filter([{}, {}])
            res6 = block.filter([{}, {}])
            res7 = block.filter(targdict={})
            res8 = block.filter(targdict=[])
            res9 = block.filter(targdict=[{}])
            res10 = block.filter(targdict=[{}, {}])

            assert_same_sub_schema(res1, targ)
            assert_same_sub_schema(res2, targ)
            assert_same_sub_schema(res3, targ)
            assert_same_sub_schema(res4, targ)
            assert_same_sub_schema(res5, targ)
            assert_same_sub_schema(res6, targ)
            assert_same_sub_schema(res7, targ)
            assert_same_sub_schema(res8, targ)
            assert_same_sub_schema(res9, targ)
            assert_same_sub_schema(res10, targ)

    def test__filter_annotation_single(self):
        block = simple_block()

        targ = [block.segments[0].analogsignals[0], block.segments[0].spiketrains[1], block.segments[1].events[0]]

        res0 = block.filter(thing="wotsit")
        res1 = block.filter({"thing": "wotsit"})
        res2 = block.filter(targdict={"thing": "wotsit"})
        res3 = block.filter([{"thing": "wotsit"}])
        res4 = block.filter(targdict=[{"thing": "wotsit"}])

        self.assertEqual(res0, targ)
        assert_same_sub_schema(res0, targ)
        assert_same_sub_schema(res1, targ)
        assert_same_sub_schema(res2, targ)
        assert_same_sub_schema(res3, targ)
        assert_same_sub_schema(res4, targ)

    def test__filter_single_annotation_nores(self):
        block = simple_block()

        res0 = block.filter(j=5)
        res1 = block.filter({"j": 5})
        res2 = block.filter(targdict={"j": 5})
        res3 = block.filter([{"j": 5}])
        res4 = block.filter(targdict=[{"j": 5}])

        self.assertEqual(len(res0), 0)
        self.assertEqual(len(res1), 0)
        self.assertEqual(len(res2), 0)
        self.assertEqual(len(res3), 0)
        self.assertEqual(len(res4), 0)

    def test__filter_attribute_single(self):
        block = simple_block()

        targ = [block.segments[1].analogsignals[0], block.segments[1].irregularlysampledsignals[0]]

        name = targ[0].name
        res0 = block.filter(name=name)
        res1 = block.filter({"name": name})
        res2 = block.filter(targdict={"name": name})

        assert_same_sub_schema(res0, targ)
        assert_same_sub_schema(res1, targ)
        assert_same_sub_schema(res2, targ)

    def test__filter_attribute_single_nores(self):
        block = simple_block()

        name = "potato"
        res0 = block.filter(name=name)
        res1 = block.filter({"name": name})
        res2 = block.filter(targdict={"name": name})

        self.assertEqual(len(res0), 0)
        self.assertEqual(len(res1), 0)
        self.assertEqual(len(res2), 0)

    def test__filter_multi(self):

        block = simple_block()
        targ = [block.segments[1].analogsignals[0], block.segments[1].irregularlysampledsignals[0]]

        filter = {"name": targ[0].name, "thing": targ[0].annotations["thing"]}

        res0 = block.filter(**filter)
        res1 = block.filter(filter)
        res2 = block.filter(targdict=filter)

        assert_same_sub_schema(res0, targ)
        assert_same_sub_schema(res1, targ)
        assert_same_sub_schema(res2, targ)

    def test__filter_multi_nores(self):
        block = simple_block()
        targ = []

        filter = {"name": "carrot", "thing": "another thing"}

        res0 = block.filter(**filter)
        res1 = block.filter(filter)
        res2 = block.filter(targdict=filter)

        assert_same_sub_schema(res0, targ)
        assert_same_sub_schema(res1, targ)
        assert_same_sub_schema(res2, targ)

    def test__filter_no_annotation_but_object(self):
        block = simple_block()
        targ = []
        for seg in block.segments:
            targ.extend(seg.events)
        res = block.filter(objects=Event)
        assert_same_sub_schema(res, targ)

        targ = []
        for seg in block.segments:
            targ.extend(seg.analogsignals)
        res = block.filter(objects=AnalogSignal)
        assert_same_sub_schema(res, targ)

        targ = []
        for seg in block.segments:
            targ.extend(seg.analogsignals)
            targ.extend(seg.events)
        res = block.filter(objects=[AnalogSignal, Event])
        assert_same_sub_schema(res, targ)

    def test__filter_single_annotation_obj_single(self):
        block = simple_block()
        targ = [block.segments[0].analogsignals[1]]

        res0 = block.filter(thing="frooble", objects="AnalogSignal")
        res1 = block.filter(thing="frooble", objects=AnalogSignal)
        res2 = block.filter(thing="frooble", objects=["AnalogSignal"])
        res3 = block.filter(thing="frooble", objects=[AnalogSignal])

        assert_same_sub_schema(res0, targ)
        assert_same_sub_schema(res1, targ)
        assert_same_sub_schema(res2, targ)
        assert_same_sub_schema(res3, targ)

    def test__filter_single_annotation_norecur(self):
        block = simple_block()
        targ = []
        res0 = block.filter(thing="frooble", recursive=False)
        assert_same_sub_schema(res0, targ)

    def test__filter_single_attribute_norecur(self):
        block = simple_block()
        targ = []
        res0 = block.filter(name=block.segments[1].analogsignals[0].name, recursive=False)
        assert_same_sub_schema(res0, targ)

    def test__filter_single_annotation_nodata(self):
        block = simple_block()
        targ = []
        res0 = block.filter(thing="frooble", data=False)
        assert_same_sub_schema(res0, targ)

    def test__filter_single_attribute_nodata(self):
        block = simple_block()
        targ = []
        res0 = block.filter(name=block.segments[0].analogsignals[0], data=False)
        assert_same_sub_schema(res0, targ)

    def test__filter_single_annotation_nodata_norecur(self):
        block = simple_block()
        targ = []
        res0 = block.filter(thing="frooble", data=False, recursive=False)
        assert_same_sub_schema(res0, targ)

    def test__filter_single_attribute_nodata_norecur(self):
        block = simple_block()
        targ = []
        res0 = block.filter(name=block.segments[0].analogsignals[0], data=False, recursive=False)
        assert_same_sub_schema(res0, targ)

    def test__filter_single_annotation_container(self):
        block = simple_block()

        targ = [block.segments[1].analogsignals[0], block.segments[1].irregularlysampledsignals[0], block.segments[1]]

        res0 = block.filter(thing="amajig", container=True)

        self.assertEqual(res0, targ)
        assert_same_sub_schema(res0, targ)

    def test__filter_single_attribute_container_data(self):
        block = simple_block()
        targ = [block.segments[1]]
        res0 = block.filter(name=targ[0].name, container=True)
        assert_same_sub_schema(res0, targ)

    def test__filter_single_annotation_container_norecur(self):
        block = simple_block()
        targ = [block.segments[1]]
        res0 = block.filter(thing="amajig", container=True, recursive=False)

        self.assertEqual(res0, targ)
        assert_same_sub_schema(res0, targ)

    def test__filter_single_attribute_container_norecur(self):
        block = simple_block()
        targ = [block.segments[1]]
        res0 = block.filter(name=targ[0].name, container=True, recursive=False)

        self.assertEqual(res0, targ)
        assert_same_sub_schema(res0, targ)

    def test__filter_single_attribute_container_norecur_nores(self):
        block = simple_block()
        targ = []
        res0 = block.filter(name="penguin", container=True, recursive=False)
        assert_same_sub_schema(res0, targ)

    def test__filter_single_annotation_nodata_container(self):
        block = simple_block()

        targ = [block.segments[1]]

        res0 = block.filter(thing="amajig", container=True, data=False)

        self.assertEqual(res0, targ)
        assert_same_sub_schema(res0, targ)

    def test__filter_single_attribute_nodata_container_nores(self):
        block = simple_block()
        targ = []
        res0 = block.filter(name="narwhal", data=False, container=True)
        assert_same_sub_schema(res0, targ)

    # def test__filter_single_annotation_nodata_container_norecur(self):
    #     targ = [self.segs1[1], self.chxs1[1]]
    #     res0 = block.filter(j=1,
    #                                data=False, container=True,
    #                                recursive=False)
    #     assert_same_sub_schema(res0, targ)

    # def test__filter_single_attribute_nodata_container_norecur(self):
    #     targ = [self.segs1[0]]
    #     res0 = block.filter(name=self.segs1[0].name,
    #                                data=False, container=True,
    #                                recursive=False)
    #     assert_same_sub_schema(res0, targ)

    def test__filter_single_attribute_nodata_container_norecur_nores(self):
        block = simple_block()
        targ = []
        res0 = block.filter(name="puffin", data=False, container=True, recursive=False)
        assert_same_sub_schema(res0, targ)

    def test__filterdata_multi(self):
        block = simple_block()
        targ = [block.segments[1].analogsignals[0], block.segments[1].irregularlysampledsignals[0], block.segments[1]]
        data = block.children_recur

        filter = {"name": targ[0].name, "thing": targ[0].annotations["thing"]}

        res0 = filterdata(data, **filter)
        res1 = filterdata(data, filter)
        res2 = filterdata(data, targdict=filter)

        assert_same_sub_schema(res0, targ)
        assert_same_sub_schema(res1, targ)
        assert_same_sub_schema(res2, targ)

    def test__filterdata_multi_nores(self):
        block = simple_block()
        targ = []
        data = block.children_recur

        name1 = (block.segments[0].analogsignals[0].name,)

        res0 = filterdata(data, [{"thing": "a good thing"}, {}])
        res1 = filterdata(data, {}, thing="a good thing")
        res2 = filterdata(data, [{}], thing="a good thing")
        res3 = filterdata(data, name=name1, targdict={"thing": "a good thing"})
        res4 = filterdata(data, {"name": name1}, thing="a good thing")
        res5 = filterdata(data, targdict={"name": name1}, thing="a good thing")
        res12 = filterdata(data, {"name": name1}, thing="a good thing")
        res13 = filterdata(data, targdict={"name": name1}, thing="a good thing")
        res14 = filterdata(data, name=name1, targdict={"thing": "a good thing"})

        assert_same_sub_schema(res0, targ)
        assert_same_sub_schema(res1, targ)
        assert_same_sub_schema(res2, targ)
        assert_same_sub_schema(res3, targ)
        assert_same_sub_schema(res4, targ)
        assert_same_sub_schema(res5, targ)
        assert_same_sub_schema(res12, targ)
        assert_same_sub_schema(res13, targ)
        assert_same_sub_schema(res14, targ)

    # @unittest.skipUnless(HAVE_IPYTHON, "requires IPython")
    # def test__pretty(self):
    #     res = pretty(self.blk1)
    #     ann = get_annotations()
    #     ann['seed'] = self.seed1
    #     ann = pretty(ann).replace('\n ', '\n  ')
    #
    #     seg0 = pretty(self.segs1[0])
    #     seg1 = pretty(self.segs1[1])
    #     seg0 = seg0.replace('\n', '\n   ')
    #     seg1 = seg1.replace('\n', '\n   ')
    #
    #     targ = ("Block with " +
    #             ("%s segments, %s channel_indexes\n" %
    #              (len(self.segs1), len(self.chxs1))) +
    #             ("name: '%s'\ndescription: '%s'\n" % (self.blk1.name,
    #                                                   self.blk1.description)) +
    #             ("annotations: %s\n" % ann) +
    #             ("file_origin: '%s'\n" % self.blk1.file_origin) +
    #             ("file_datetime: %s\n" % repr(self.blk1.file_datetime)) +
    #             ("rec_datetime: %s\n" % repr(self.blk1.rec_datetime)) +
    #             ("index: %s\n" % self.blk1.index) +
    #
    #
    #             ("# segments (N=%s)\n" % len(self.segs1)) +
    #             ('%s: %s\n' % (0, seg0)) +
    #             ('%s: %s' % (1, seg1)))
    #
    #     self.assertEqual(res, targ)

    def test__deepcopy(self):
        blk1 = self.blocks[0]
        blk1_copy = deepcopy(blk1)

        # Check links from parents to children and object attributes
        assert_same_sub_schema(blk1_copy, blk1)

        # Check links from children to parents
        for segment in blk1_copy.segments:
            self.assertEqual(id(segment.block), id(blk1_copy))
            for sig in segment.analogsignals:
                self.assertEqual(id(sig.segment), id(segment))
            for sptr in segment.spiketrains:
                self.assertEqual(id(sptr.segment), id(segment))

    def test_segment_list(self):
        blk = Block()
        assert len(blk.segments) == 0
        blk.segments.append(Segment())
        assert len(blk.segments) == 1
        blk.segments.extend([Segment(), Segment()])
        assert len(blk.segments) == 3
        blk.segments = []
        assert len(blk.segments) == 0
        blk.segments = [Segment()]
        assert len(blk.segments) == 1
        blk.segments += [Segment(), Segment()]
        assert len(blk.segments) == 3

    def test_add(self):
        blk = self.blocks[0]
        new_blk = simple_block()
        n_groups_start = len(new_blk.groups)
        for group in blk.groups:
            assert group not in new_blk.groups
            new_blk.add(group)
            assert group in new_blk.groups
        assert len(new_blk.groups) == n_groups_start + len(blk.groups)

        n_segs_start = len(new_blk.segments)
        for seg in blk.segments:
            assert seg not in new_blk.segments
            new_blk.add(seg)
            assert seg in new_blk.segments
        assert len(new_blk.segments) == n_segs_start + len(blk.segments)

        # test adding multiple at once
        blk = self.blocks[1]
        n_groups_start = len(new_blk.groups)
        new_blk.add(*blk.groups)
        assert len(new_blk.groups) == n_groups_start + len(blk.groups)

        n_segs_start = len(new_blk.segments)
        new_blk.add(*blk.segments)
        assert len(new_blk.segments) == n_segs_start + len(blk.segments)

    def test_add_invalid_type_raises_Exception(self):
        new_blk = Block()
        self.assertRaises(TypeError, new_blk.add, random_signal())


if __name__ == "__main__":
    unittest.main()



"""
Tests of neo.io.elanio
"""

import unittest

from neo.io import ElanIO
from neo.test.iotest.common_io_test import BaseTestIO


class TestElanIO(
    BaseTestIO,
    unittest.TestCase,
):
    ioclass = ElanIO
    entities_to_download = ["elan"]
    entities_to_test = [
        "elan/File_elan_1.eeg",
    ]


if __name__ == "__main__":
    unittest.main()



"""
Tests of neo.io.medio
"""

import pathlib
import unittest

import quantities as pq
import numpy as np

from neo.io.medio import MedIO
from neo.test.iotest.common_io_test import BaseTestIO

try:
    import dhn_med_py

    HAVE_DHN_MED = True
except ImportError:
    HAVE_DHN_MED = False


# This runs standard tests, this is mandatory for all IOs
@unittest.skipUnless(HAVE_DHN_MED, "requires dhn_med_py package and all its dependencies")
class TestMedIO(
    BaseTestIO,
    unittest.TestCase,
):
    ioclass = MedIO
    entities_to_download = ["med"]
    entities_to_test = ["med/sine_waves.medd", "med/test.medd"]

    def setUp(self):

        super().setUp()
        self.dirname = self.get_local_path("med/sine_waves.medd")
        self.dirname2 = self.get_local_path("med/test.medd")
        self.password = "L2_password"

    def test_read_segment_lazy(self):

        r = MedIO(self.dirname, self.password)
        seg = r.read_segment(lazy=False)

        # There will only be one analogsignal in this reading
        self.assertEqual(len(seg.analogsignals), 1)
        # Test that the correct number of samples are read, 5760000 samps for 3 channels
        self.assertEqual(seg.analogsignals[0].shape[0], 5760000)
        self.assertEqual(seg.analogsignals[0].shape[1], 3)

        # Test the first sample value of all 3 channels, which are
        # known to be [-1, -4, -4]
        np.testing.assert_array_equal(seg.analogsignals[0][0][:3], [-1, -4, -4])

        for anasig in seg.analogsignals:
            self.assertNotEqual(anasig.size, 0)
        for st in seg.spiketrains:
            self.assertNotEqual(st.size, 0)

        # annotations
        # assert 'seg_extra_info' in seg.annotations
        assert seg.name == "Seg #0 Block #0"
        for anasig in seg.analogsignals:
            assert anasig.name is not None
        for ev in seg.events:
            assert ev.name is not None
        for ep in seg.epochs:
            assert ep.name is not None

        r.close()

    def test_read_block(self):

        r = MedIO(self.dirname, self.password)
        bl = r.read_block(lazy=True)
        self.assertTrue(bl.annotations)

        for count, seg in enumerate(bl.segments):
            assert seg.name == "Seg #" + str(count) + " Block #0"

        for anasig in seg.analogsignals:
            assert anasig.name is not None

        # Verify that the block annotations from the MED session are
        # read properly.  There are a lot of annotations, so we'll just
        # spot-check a couple of them.
        assert bl.annotations["metadata"]["recording_country"] == "United States"
        assert bl.annotations["metadata"]["AC_line_frequency"] == 60.0

        r.close()

    def test_read_segment_with_time_slice(self):
        """
        Test loading of a time slice and check resulting times
        """
        r = MedIO(self.dirname, self.password)
        seg = r.read_segment(time_slice=None)

        # spike and epoch timestamps are not being read
        self.assertEqual(len(seg.spiketrains), 0)
        self.assertEqual(len(seg.epochs), 1)
        self.assertEqual(len(seg.epochs[0]), 0)

        # Test for 180 events (1 per second for 3 minute recording)
        self.assertEqual(len(seg.events), 1)
        self.assertEqual(len(seg.events[0]), 180)

        for asig in seg.analogsignals:
            self.assertEqual(asig.shape[0], 5760000)
        n_channels = sum(a.shape[-1] for a in seg.analogsignals)
        self.assertEqual(n_channels, 3)

        t_start, t_stop = 500 * pq.ms, 800 * pq.ms
        seg = r.read_segment(time_slice=(t_start, t_stop))

        # Test that 300 ms were read, which at 32 kHz, is 9600 samples
        self.assertAlmostEqual(seg.analogsignals[0].shape[0], 9600, delta=1.0)
        # Test that it read from 3 channels
        self.assertEqual(seg.analogsignals[0].shape[1], 3)

        self.assertAlmostEqual(seg.t_start.rescale(t_start.units), t_start, delta=5.0)
        self.assertAlmostEqual(seg.t_stop.rescale(t_stop.units), t_stop, delta=5.0)

        r.close()


if __name__ == "__main__":
    unittest.main()



"""
RawIO Class for NIX files

The RawIO assumes all segments and all blocks have the same structure.
It supports all kinds of NEO objects.

Author: Chek Yin Choi, Julia Sprenger
"""

import numpy as np

from .baserawio import (
    BaseRawIO,
    _signal_channel_dtype,
    _signal_stream_dtype,
    _spike_channel_dtype,
    _event_channel_dtype,
)
from ..io.nixio import check_nix_version


# When reading metadata properties, the following keys are ignored since they
# are used to store Neo object properties.
# This dictionary is used in the _filter_properties() method.
neo_attributes = {
    "segment": ["index"],
    "analogsignal": ["units", "copy", "sampling_rate", "t_start"],
    "spiketrain": ["units", "copy", "sampling_rate", "t_start", "t_stop", "waveforms", "left_sweep"],
    "event": ["times", "labels", "units", "durations", "copy"],
}


class NIXRawIO(BaseRawIO):
    """Class for reading NIX files

    Parameters
    ----------
    filename: str, default: ''
        The nix file to be load

    Notes
    -----
    Requires the nixio package to be installed"""

    extensions = ["nix", "h5"]
    rawmode = "one-file"

    def __init__(self, filename=""):
        check_nix_version()
        BaseRawIO.__init__(self)
        self.filename = str(filename)

    def _source_name(self):
        return self.filename

    def _parse_header(self):
        import nixio

        self.file = nixio.File.open(self.filename, nixio.FileMode.ReadOnly)
        signal_channels = []
        anasig_ids = {0: []}  # ids of analogsignals by segment
        stream_ids = []
        for bl in self.file.blocks:
            for seg in bl.groups:
                for da_idx, da in enumerate(seg.data_arrays):
                    if da.type == "neo.analogsignal":
                        chan_id = da_idx
                        ch_name = da.metadata["neo_name"]
                        units = str(da.unit)
                        dtype = str(da.dtype)
                        sr = 1 / da.dimensions[0].sampling_interval
                        anasig_id = da.name.split(".")[-2]
                        if anasig_id not in anasig_ids[0]:
                            anasig_ids[0].append(anasig_id)
                        stream_id = anasig_ids[0].index(anasig_id)
                        if stream_id not in stream_ids:
                            stream_ids.append(stream_id)
                        gain = 1
                        offset = 0.0
                        signal_channels.append((ch_name, chan_id, sr, dtype, units, gain, offset, stream_id))
                # only read structure of first segment and assume the same
                # across segments
                break
            break
        signal_channels = np.array(signal_channels, dtype=_signal_channel_dtype)
        signal_streams = np.zeros(len(stream_ids), dtype=_signal_stream_dtype)
        signal_streams["id"] = stream_ids
        signal_streams["name"] = ""

        spike_channels = []
        unit_name = ""
        unit_id = ""
        for bl in self.file.blocks:
            seg_groups = [g for g in bl.groups if g.type == "neo.segment"]

            for seg in seg_groups:
                for mt in seg.multi_tags:
                    if mt.type == "neo.spiketrain":
                        unit_name = mt.metadata["neo_name"]
                        unit_id = mt.id
                        wf_left_sweep = 0
                        wf_units = None
                        wf_sampling_rate = 0
                        if mt.features:
                            wf = mt.features[0].data
                            wf_units = wf.unit
                            dim = wf.dimensions[2]
                            interval = dim.sampling_interval
                            wf_sampling_rate = 1 / interval
                            if wf.metadata:
                                wf_left_sweep = wf.metadata["left_sweep"]
                        wf_gain = 1
                        wf_offset = 0.0
                        spike_channels.append(
                            (unit_name, unit_id, wf_units, wf_gain, wf_offset, wf_left_sweep, wf_sampling_rate)
                        )
                break
            break
        spike_channels = np.array(spike_channels, dtype=_spike_channel_dtype)

        event_channels = []
        event_count = 0
        epoch_count = 0
        for bl in self.file.blocks:
            seg_groups = [g for g in bl.groups if g.type == "neo.segment"]
            for seg in seg_groups:
                for mt in seg.multi_tags:
                    if mt.type == "neo.event":
                        ev_name = mt.metadata["neo_name"]
                        ev_id = event_count
                        event_count += 1
                        ev_type = "event"
                        event_channels.append((ev_name, ev_id, ev_type))
                    if mt.type == "neo.epoch":
                        ep_name = mt.metadata["neo_name"]
                        ep_id = epoch_count
                        epoch_count += 1
                        ep_type = "epoch"
                        event_channels.append((ep_name, ep_id, ep_type))
                break
            break
        event_channels = np.array(event_channels, dtype=_event_channel_dtype)

        self.da_list = {"blocks": []}
        for block_index, blk in enumerate(self.file.blocks):
            seg_groups = [g for g in blk.groups if g.type == "neo.segment"]
            d = {"segments": []}
            self.da_list["blocks"].append(d)
            for seg_index, seg in enumerate(seg_groups):
                d = {"signals": []}
                self.da_list["blocks"][block_index]["segments"].append(d)
                size_list = []
                data_list = []
                da_name_list = []
                for da in seg.data_arrays:
                    if da.type == "neo.analogsignal":
                        size_list.append(da.size)
                        data_list.append(da)
                        da_name_list.append(da.metadata["neo_name"])
                block = self.da_list["blocks"][block_index]
                segment = block["segments"][seg_index]
                segment["data_size"] = size_list
                segment["data"] = data_list
                segment["ch_name"] = da_name_list

        self.unit_list = {"blocks": []}
        for block_index, blk in enumerate(self.file.blocks):
            seg_groups = [g for g in blk.groups if g.type == "neo.segment"]
            d = {"segments": []}
            self.unit_list["blocks"].append(d)
            for seg_index, seg in enumerate(seg_groups):
                d = {"spiketrains": [], "spiketrains_id": [], "spiketrains_unit": []}
                self.unit_list["blocks"][block_index]["segments"].append(d)
                st_idx = 0
                for st in seg.multi_tags:
                    d = {"waveforms": []}
                    block = self.unit_list["blocks"][block_index]
                    segment = block["segments"][seg_index]
                    segment["spiketrains_unit"].append(d)
                    if st.type == "neo.spiketrain":
                        segment["spiketrains"].append(st.positions)
                        segment["spiketrains_id"].append(st.id)
                        wftypestr = "neo.waveforms"
                        if st.features and st.features[0].data.type == wftypestr:
                            waveforms = st.features[0].data
                            stdict = segment["spiketrains_unit"][st_idx]
                            if waveforms:
                                stdict["waveforms"] = waveforms
                            else:
                                stdict["waveforms"] = None
                            # assume one spiketrain one waveform
                            st_idx += 1

        self.header = {}
        self.header["nb_block"] = len(self.file.blocks)
        self.header["nb_segment"] = [len(seg_groups) for bl in self.file.blocks]
        self.header["signal_streams"] = signal_streams
        self.header["signal_channels"] = signal_channels
        self.header["spike_channels"] = spike_channels
        self.header["event_channels"] = event_channels

        self._generate_minimal_annotations()
        for blk_idx, blk in enumerate(self.file.blocks):
            seg_groups = [g for g in blk.groups if g.type == "neo.segment"]
            bl_ann = self.raw_annotations["blocks"][blk_idx]
            props = blk.metadata.inherited_properties()
            bl_ann.update(self._filter_properties(props, "block"))
            for grp_idx, group in enumerate(seg_groups):
                seg_ann = bl_ann["segments"][grp_idx]
                props = group.metadata.inherited_properties()
                seg_ann.update(self._filter_properties(props, "segment"))

                sp_idx = 0
                ev_idx = 0
                for mt in group.multi_tags:
                    if mt.type == "neo.spiketrain" and seg_ann["spikes"]:
                        st_ann = seg_ann["spikes"][sp_idx]
                        props = mt.metadata.inherited_properties()
                        st_ann.update(self._filter_properties(props, "spiketrain"))
                        sp_idx += 1
                    # if order is preserving, the annotations
                    # should go to the right place, need test
                    if mt.type == "neo.event" or mt.type == "neo.epoch":
                        if seg_ann["events"] != []:
                            event_ann = seg_ann["events"][ev_idx]
                            props = mt.metadata.inherited_properties()
                            event_ann.update(self._filter_properties(props, "event"))
                            ev_idx += 1

                # adding array annotations to analogsignals
                annotated_anasigs = []
                sig_ann = seg_ann["signals"]
                # this implementation relies on analogsignals always being
                # stored in the same stream order across segments
                stream_id = 0
                for da_idx, da in enumerate(group.data_arrays):
                    if da.type != "neo.analogsignal":
                        continue
                    anasig_id = da.name.split(".")[-2]
                    # skip already annotated signals as each channel already
                    # contains the complete set of annotations and
                    # array_annotations
                    if anasig_id in annotated_anasigs:
                        continue
                    annotated_anasigs.append(anasig_id)

                    # collect annotation properties
                    props = [p for p in da.metadata.props if p.type != "ARRAYANNOTATION"]
                    props_dict = self._filter_properties(props, "analogsignal")
                    sig_ann[stream_id].update(props_dict)

                    # collect array annotation properties
                    props = [p for p in da.metadata.props if p.type == "ARRAYANNOTATION"]
                    props_dict = self._filter_properties(props, "analogsignal")
                    sig_ann[stream_id]["__array_annotations__"].update(props_dict)

                    stream_id += 1

    def _segment_t_start(self, block_index, seg_index):
        t_start = 0
        for mt in self.file.blocks[block_index].groups[seg_index].multi_tags:
            if mt.type == "neo.spiketrain":
                t_start = mt.metadata["t_start"]
        return t_start

    def _segment_t_stop(self, block_index, seg_index):
        t_stop = 0
        for mt in self.file.blocks[block_index].groups[seg_index].multi_tags:
            if mt.type == "neo.spiketrain":
                t_stop = mt.metadata["t_stop"]
        return t_stop

    def _get_signal_size(self, block_index, seg_index, stream_index):
        stream_id = self.header["signal_streams"][stream_index]["id"]
        keep = self.header["signal_channels"]["stream_id"] == stream_id
        (channel_indexes,) = np.nonzero(keep)
        ch_idx = channel_indexes[0]
        block = self.da_list["blocks"][block_index]
        segment = block["segments"][seg_index]
        size = segment["data_size"][ch_idx]
        return size  # size is per signal, not the sum of all channel_indexes

    def _get_signal_t_start(self, block_index, seg_index, stream_index):
        stream_id = self.header["signal_streams"][stream_index]["id"]
        keep = self.header["signal_channels"]["stream_id"] == stream_id
        (channel_indexes,) = np.nonzero(keep)
        ch_idx = channel_indexes[0]
        block = self.file.blocks[block_index]
        das = [da for da in block.groups[seg_index].data_arrays]
        da = das[ch_idx]
        sig_t_start = float(da.metadata["t_start"])
        return sig_t_start  # assume same group_id always same t_start

    def _get_analogsignal_chunk(self, block_index, seg_index, i_start, i_stop, stream_index, channel_indexes):
        stream_id = self.header["signal_streams"][stream_index]["id"]
        keep = self.header["signal_channels"]["stream_id"] == stream_id
        (global_channel_indexes,) = np.nonzero(keep)
        if channel_indexes is not None:
            global_channel_indexes = global_channel_indexes[channel_indexes]

        if i_start is None:
            i_start = 0
        if i_stop is None:
            i_stop = self.get_signal_size(block_index, seg_index, stream_index)

        raw_signals_list = []
        da_list = self.da_list["blocks"][block_index]["segments"][seg_index]
        for idx in global_channel_indexes:
            da = da_list["data"][idx]
            raw_signals_list.append(da[i_start:i_stop])

        raw_signals = np.array(raw_signals_list)
        raw_signals = np.transpose(raw_signals)
        return raw_signals

    def _spike_count(self, block_index, seg_index, unit_index):
        count = 0
        head_id = self.header["spike_channels"][unit_index][1]
        for mt in self.file.blocks[block_index].groups[seg_index].multi_tags:
            for src in mt.sources:
                if mt.type == "neo.spiketrain" and [src.type == "neo.unit"]:
                    if head_id == src.id:
                        return len(mt.positions)
        return count

    def _get_spike_timestamps(self, block_index, seg_index, unit_index, t_start, t_stop):
        block = self.unit_list["blocks"][block_index]
        segment = block["segments"][seg_index]
        spike_dict = segment["spiketrains"]
        spike_timestamps = spike_dict[unit_index]
        spike_timestamps = np.transpose(spike_timestamps)

        if t_start is not None or t_stop is not None:
            lim0 = t_start
            lim1 = t_stop
            mask = (spike_timestamps >= lim0) & (spike_timestamps <= lim1)
            spike_timestamps = spike_timestamps[mask]
        return spike_timestamps

    def _rescale_spike_timestamp(self, spike_timestamps, dtype):
        spike_times = spike_timestamps.astype(dtype)
        return spike_times

    def _get_spike_raw_waveforms(self, block_index, seg_index, unit_index, t_start, t_stop):
        # this must return a 3D numpy array (nb_spike, nb_channel, nb_sample)
        seg = self.unit_list["blocks"][block_index]["segments"][seg_index]
        waveforms = seg["spiketrains_unit"][unit_index]["waveforms"]
        if not waveforms:
            return None
        raw_waveforms = np.array(waveforms)

        if t_start is not None:
            lim0 = t_start
            mask = raw_waveforms >= lim0
            # use nan to keep the shape
            raw_waveforms = np.where(mask, raw_waveforms, np.nan)
        if t_stop is not None:
            lim1 = t_stop
            mask = raw_waveforms <= lim1
            raw_waveforms = np.where(mask, raw_waveforms, np.nan)
        return raw_waveforms

    def _event_count(self, block_index, seg_index, event_channel_index):
        event_count = 0
        segment = self.file.blocks[block_index].groups[seg_index]
        for event in segment.multi_tags:
            if event.type == "neo.event" or event.type == "neo.epoch":
                if event_count == event_channel_index:
                    return len(event.positions)
                else:
                    event_count += 1
        return event_count

    def _get_event_timestamps(self, block_index, seg_index, event_channel_index, t_start, t_stop):
        timestamp = []
        labels = []
        durations = None
        if event_channel_index is None:
            raise IndexError
        for mt in self.file.blocks[block_index].groups[seg_index].multi_tags:
            if mt.type == "neo.event" or mt.type == "neo.epoch":
                labels.append(mt.positions.dimensions[0].labels)
                po = mt.positions
                if po.type == "neo.event.times" or po.type == "neo.epoch.times":
                    timestamp.append(po)
                channel = self.header["event_channels"][event_channel_index]
                if channel["type"] == b"epoch" and mt.extents:
                    if mt.extents.type == "neo.epoch.durations":
                        durations = np.array(mt.extents)
                        break
        timestamp = timestamp[event_channel_index][:]
        timestamp = np.array(timestamp, dtype="float")
        labels = labels[event_channel_index][:]
        labels = np.array(labels, dtype="U")
        if t_start is not None:
            keep = timestamp >= t_start
            timestamp, labels = timestamp[keep], labels[keep]

        if t_stop is not None:
            keep = timestamp <= t_stop
            timestamp, labels = timestamp[keep], labels[keep]
        return timestamp, durations, labels  # only the first fits in rescale

    def _rescale_event_timestamp(self, event_timestamps, dtype, event_channel_index):
        ev_unit = ""
        for mt in self.file.blocks[0].groups[0].multi_tags:
            if mt.type == "neo.event":
                ev_unit = mt.positions.unit
                break
        if ev_unit == "ms":
            event_timestamps /= 1000
        event_times = event_timestamps.astype(dtype)
        # supposing unit is second, other possibilities maybe mS microS...
        return event_times  # return in seconds

    def _rescale_epoch_duration(self, raw_duration, dtype, event_channel_index):
        ep_unit = ""
        for mt in self.file.blocks[0].groups[0].multi_tags:
            if mt.type == "neo.epoch":
                ep_unit = mt.positions.unit
                break
        if ep_unit == "ms":
            raw_duration /= 1000
        durations = raw_duration.astype(dtype)
        # supposing unit is second, other possibilities maybe mS microS...
        return durations  # return in seconds

    def _filter_properties(self, properties, neo_type):
        """
        Takes a collection of NIX metadata properties and the name of a Neo
        type and returns a dictionary representing the Neo object annotations.
        Properties that represent the attributes of the Neo object type are
        filtered, based on the global 'neo_attributes' dictionary.
        """
        annotations = dict()
        attrs = neo_attributes.get(neo_type, list())
        for prop in properties:
            # filter neo_name explicitly
            if not (prop.name in attrs or prop.name == "neo_name"):
                values = prop.values
                if len(values) == 1:
                    values = values[0]
                annotations[str(prop.name)] = values
        return annotations



"""
Working with RegionOfInterest objects
=====================================

"""

#################################################################
# Import our packages
# We can import a variety of neo objects from neo.core and since an ImageSquence
# also uses units let's import quantities which is the "units" library
# that neo uses under the hood

import matplotlib.pyplot as plt
import numpy as np
from neo.core import CircularRegionOfInterest, RectangularRegionOfInterest, PolygonRegionOfInterest, ImageSequence
from numpy.random import rand
import random
import quantities as pq

##################################################################
# First we create our image_sequence. Let's generate some data
# In this simulated dataset we will just make an image that is
# 100x100 pixels and then we will make 50 frames of this image
# finally we will fill with random values for the data

l = []
for frame in range(50):
    l.append([])
    for y in range(100):
        l[frame].append([])
        for x in range(100):
            l[frame][y].append(random.randint(0, 50))

# make an ImageSquence in Neo
image_seq = ImageSequence(l, sampling_rate=500 * pq.Hz, spatial_scale="m", units="V")


#################################################################
# Now we will write a function for plotting an roi
def plot_roi(roi, shape):
    img = rand(120, 100)
    pir = np.array(roi.pixels_in_region()).T
    img[pir[1], pir[0]] = 5

    plt.imshow(img, cmap="gray_r")
    plt.clim(0, 5)

    ax = plt.gca()
    ax.add_artist(shape)


################################################################################
# Finally we will plot each roi to demonstrate how we could high regions of interest

# First a nice circle
roi = CircularRegionOfInterest(image_sequence=image_seq, x=50.3, y=50.8, radius=30.2)
shape = plt.Circle(roi.centre, roi.radius, color="r", fill=False)
plt.subplot(1, 3, 1)
plot_roi(roi, shape)

# Next a rectangle
roi = RectangularRegionOfInterest(image_sequence=image_seq, x=50.3, y=40.2, width=40.1, height=50.3)
shape = plt.Rectangle((roi.x - roi.width / 2.0, roi.y - roi.height / 2.0), roi.width, roi.height, color="r", fill=False)
plt.subplot(1, 3, 2)
plot_roi(roi, shape)

# Finally we can make a polygon (in this case a triangle)
roi = PolygonRegionOfInterest(image_seq, (20.3, 30.2), (80.7, 30.1), (55.2, 59.4))
shape = plt.Polygon(np.array(roi.vertices), closed=True, color="r", fill=False)
plt.subplot(1, 3, 3)
plot_roi(roi, shape)

plt.tight_layout()
plt.show()



# -*- coding: utf-8 -*-
"""
Tests of neo.io.stimfitio
"""

import sys

import unittest

from neo.io import StimfitIO
from neo.test.iotest.common_io_test import BaseTestIO

try:
    import stfio
except Exception:
    HAS_STFIO = False
else:
    HAS_STFIO = True


@unittest.skipIf(sys.version_info[0] > 2, "not Python 3 compatible")
@unittest.skipUnless(HAS_STFIO, "requires stfio")
class TestStimfitIO(BaseTestIO, unittest.TestCase):
    ioclass = StimfitIO
    entities_to_download = ["stimfit"]
    entities_to_test = [
        "stimfit/File_stimfit_1.h5",
        "stimfit/File_stimfit_2.h5",
        "stimfit/File_stimfit_3.h5",
        "stimfit/File_stimfit_4.h5",
        "stimfit/File_stimfit_5.h5",
        "stimfit/File_stimfit_6.h5",
    ]


if __name__ == "__main__":
    unittest.main()



"""
Tests of neo.rawio.bci2000rawio
"""

import unittest

from neo.rawio.bci2000rawio import BCI2000RawIO
from neo.test.rawiotest.common_rawio_test import BaseTestRawIO


class TestBCI2000RawIO(
    BaseTestRawIO,
    unittest.TestCase,
):
    rawioclass = BCI2000RawIO

    entities_to_download = ["bci2000/eeg1_1.dat", "bci2000/eeg1_2.dat", "bci2000/eeg1_3.dat"]

    entities_to_download = [
        "bci2000",
    ]


if __name__ == "__main__":
    unittest.main()



"""
Tests of neo.io.exampleio
"""

import pathlib
import unittest

from neo.io.exampleio import ExampleIO  # , HAVE_SCIPY
from neo.test.iotest.common_io_test import BaseTestIO
from neo.test.iotest.tools import get_test_file_full_path
from neo.io.proxyobjects import AnalogSignalProxy, SpikeTrainProxy, EventProxy, EpochProxy
from neo import AnalogSignal, SpikeTrain

import quantities as pq
import numpy as np


# This run standart tests, this is mandatory for all IO
class TestExampleIO(
    BaseTestIO,
    unittest.TestCase,
):
    ioclass = ExampleIO
    entities_to_download = []
    entities_to_test = [
        "fake1.fake",
        "fake2.fake",
    ]

    def setUp(self):
        super().setUp()
        # ensure fake test files exist before running common tests
        for entity in self.entities_to_test:
            full_path = get_test_file_full_path(self.ioclass, filename=entity, directory=self.local_test_dir)
            pathlib.Path(full_path).touch()

    def tearDown(self) -> None:
        super().tearDown()
        for entity in self.entities_to_test:
            full_path = get_test_file_full_path(self.ioclass, filename=entity, directory=self.local_test_dir)
            pathlib.Path(full_path).unlink(missing_ok=True)


# This is the minimal variables that are required
# to run the common IO tests.  IO specific tests
# can be added here and will be run automatically
# in addition to the common tests.
class Specific_TestExampleIO(unittest.TestCase):
    def test_read_segment_lazy(self):
        r = ExampleIO(filename=None)
        seg = r.read_segment(lazy=True)
        for ana in seg.analogsignals:
            assert isinstance(ana, AnalogSignalProxy)
            ana = ana.load()
            assert isinstance(ana, AnalogSignal)
        for st in seg.spiketrains:
            assert isinstance(st, SpikeTrainProxy)
            st = st.load()
            assert isinstance(st, SpikeTrain)

        seg = r.read_segment(lazy=False)
        for anasig in seg.analogsignals:
            assert isinstance(ana, AnalogSignal)
            self.assertNotEqual(anasig.size, 0)
        for st in seg.spiketrains:
            assert isinstance(st, SpikeTrain)
            self.assertNotEqual(st.size, 0)

        # annotations
        assert "seg_extra_info" in seg.annotations
        assert seg.name == "Seg #0 Block #0"
        for anasig in seg.analogsignals:
            assert anasig.name is not None
        for st in seg.spiketrains:
            assert st.name is not None
        for ev in seg.events:
            assert ev.name is not None
        for ep in seg.epochs:
            assert ep.name is not None

    def test_read_block(self):
        r = ExampleIO(filename=None)
        bl = r.read_block(lazy=True)
        # assert len(bl.list_units) == 3
        # assert len(bl.channel_indexes) == 1 + 1  # signals grouped + units grouped

    def test_read_segment_with_time_slice(self):
        r = ExampleIO(filename=None)
        seg = r.read_segment(time_slice=None)
        shape_full = seg.analogsignals[0].shape
        spikes_full = seg.spiketrains[0]
        event_full = seg.events[0]

        t_start, t_stop = 260 * pq.ms, 1.854 * pq.s
        seg = r.read_segment(time_slice=(t_start, t_stop))
        shape_slice = seg.analogsignals[0].shape
        spikes_slice = seg.spiketrains[0]
        event_slice = seg.events[0]

        assert shape_full[0] > shape_slice[0]

        assert spikes_full.size > spikes_slice.size
        assert np.all(spikes_slice >= t_start)
        assert np.all(spikes_slice <= t_stop)
        assert spikes_slice.t_start == t_start
        assert spikes_slice.t_stop == t_stop

        assert event_full.size > event_slice.size
        assert np.all(event_slice.times >= t_start)
        assert np.all(event_slice.times <= t_stop)


if __name__ == "__main__":
    unittest.main()



import unittest
import numpy as np

from neo.rawio.medrawio import MedRawIO

from neo.test.rawiotest.common_rawio_test import BaseTestRawIO

try:
    import dhn_med_py

    HAVE_DHN_MED = True
except ImportError:
    HAVE_DHN_MED = False


# This runs standard tests, this is mandatory for all IOs
@unittest.skipUnless(HAVE_DHN_MED, "requires dhn_med_py package and all its dependencies")
class TestMedRawIO(
    BaseTestRawIO,
    unittest.TestCase,
):
    rawioclass = MedRawIO
    entities_to_download = ["med"]
    entities_to_test = ["med/sine_waves.medd", "med/test.medd"]

    def test_close(self):

        filename = self.get_local_path("med/sine_waves.medd")

        raw_io1 = MedRawIO(filename, password="L2_password")
        raw_io1.parse_header()
        raw_io1.close()

        raw_io2 = MedRawIO(filename, password="L2_password")
        raw_io2.parse_header()
        raw_io2.close()

    def test_scan_med_directory(self):

        filename = self.get_local_path("med/sine_waves.medd")

        rawio = MedRawIO(filename, password="L2_password")
        rawio.parse_header()

        # Test that correct metadata and boundaries are extracted
        # from the MED session.  We know the correct answers since
        # we generated the test files.
        self.assertEqual(rawio.signal_streams_count(), 1)
        self.assertEqual(rawio._segment_t_start(0, 0), 0)
        self.assertEqual(rawio._segment_t_stop(0, 0), 180)

        # Verify it found all 3 channels
        self.assertEqual(rawio.num_channels_in_session, 3)
        self.assertEqual(rawio.header["signal_channels"].size, 3)

        # Verify if found the names of the 3 channels
        self.assertEqual(rawio.header["signal_channels"][0][0], "CSC_0001")
        self.assertEqual(rawio.header["signal_channels"][1][0], "CSC_0002")
        self.assertEqual(rawio.header["signal_channels"][2][0], "CSC_0003")

        # Read first 3 seconds of data from all channels
        raw_chunk = rawio.get_analogsignal_chunk(
            block_index=0, seg_index=0, i_start=0, i_stop=96000, stream_index=0, channel_indexes=None
        )

        # Test the first sample value of all 3 channels, which are
        # known to be [-1, -4, -4]
        np.testing.assert_array_equal(raw_chunk[0][:3], [-1, -4, -4])

        # Read 1 second of data from the second channel
        raw_chunk = rawio.get_analogsignal_chunk(
            block_index=0, seg_index=0, i_start=0, i_stop=32000, stream_index=0, channel_indexes=[1]
        )

        # Test known first sample of second channel: [-4]
        self.assertEqual(raw_chunk[0][0], -4)

        rawio.close()

        # Test on second test dataset, test.medd.
        filename = self.get_local_path("med/test.medd")

        rawio = MedRawIO(filename, password="L2_password")
        rawio.parse_header()

        # Test that correct metadata and boundaries are extracted
        # from the MED session.  We know the correct answers since
        # we generated the test files.

        # For this dataset, there are 3 continuous data ranges, with
        # approximately 10 seconds between the ranges.
        # There are 3 channels, two with a frequency of 1000 Hz and one
        # with a frequency of 5000 Hz.
        self.assertEqual(rawio.signal_streams_count(), 2)

        # Segment 0
        self.assertEqual(rawio._segment_t_start(0, 0), 0)
        self.assertEqual(rawio._segment_t_stop(0, 0), 39.8898)
        # Segment 1
        self.assertEqual(rawio._segment_t_start(0, 1), 50.809826)
        self.assertEqual(rawio._segment_t_stop(0, 1), 87.337646)
        # Segment 2
        self.assertEqual(rawio._segment_t_start(0, 2), 97.242057)
        self.assertEqual(rawio._segment_t_stop(0, 2), 180.016702)

        # Verify it found all 3 channels.
        self.assertEqual(rawio.num_channels_in_session, 3)
        self.assertEqual(rawio.header["signal_channels"].size, 3)

        # Verity if found the names of the 3 channels
        self.assertEqual(rawio.header["signal_channels"][0][0], "5k_ch1")
        self.assertEqual(rawio.header["signal_channels"][1][0], "1k_ch1")
        self.assertEqual(rawio.header["signal_channels"][2][0], "1k_ch2")

        # Read first 3 seconds of data from the first channel (5k_ch1)
        raw_chunk = rawio.get_analogsignal_chunk(
            block_index=0, seg_index=0, i_start=0, i_stop=15000, stream_index=0, channel_indexes=None
        )

        # Test the first three sample values returned, which are
        # known to be [-80, -79, -78]
        self.assertEqual(raw_chunk[0][0], -80)
        self.assertEqual(raw_chunk[1][0], -79)
        self.assertEqual(raw_chunk[2][0], -78)

        # Read first 3 seconds of data from the second channel and third
        # channels (1k_ch1 and 1k_ch2)
        raw_chunk = rawio.get_analogsignal_chunk(
            block_index=0, seg_index=0, i_start=0, i_stop=3000, stream_index=1, channel_indexes=None
        )

        # Test first sample returned of both channels, which are known
        # to be [-79, -80]
        np.testing.assert_array_equal(raw_chunk[0][:2], [-79, -80])

        # Read first 3 seconds of data from the second segment of the first channel (5k_ch1)
        raw_chunk = rawio.get_analogsignal_chunk(
            block_index=0, seg_index=1, i_start=0, i_stop=15000, stream_index=0, channel_indexes=None
        )

        # Test the first three sample values returned, which are
        # known to be [22, 23, 24]
        self.assertEqual(raw_chunk[0][0], 22)
        self.assertEqual(raw_chunk[1][0], 23)
        self.assertEqual(raw_chunk[2][0], 24)

        self.assertEqual(len(rawio.header["event_channels"]), 2)

        # Verify that there are 5 events in the dataset.
        # They are 3 "Note" type events, and 2 "NlxP", or neuralynx, type events.
        # The first segment has one event, and the second and third segments
        # each have 2 events.
        self.assertEqual(rawio.event_count(0, 0, 0), 1)
        self.assertEqual(rawio.event_count(0, 1, 0), 2)
        self.assertEqual(rawio.event_count(0, 2, 0), 2)

        # Get array of all events in first segment of data
        events = rawio._get_event_timestamps(0, 0, 0, rawio._segment_t_start(0, 0), rawio._segment_t_stop(0, 0))
        # Make sure it read 1 event
        self.assertEqual(len(events[0]), 1)

        # Get array of all events in second segment of data
        events = rawio._get_event_timestamps(0, 1, 0, rawio._segment_t_start(0, 1), rawio._segment_t_stop(0, 1))
        # Make sure it read 2 events
        self.assertEqual(len(events[0]), 2)

        # Verify the first event of the second segment is a Neuralynx type event, with correct time
        self.assertEqual(events[2][0][:4], "NlxP")
        self.assertEqual(events[0][0], 51.703509)

        # Get array of all events in third segment of data
        events = rawio._get_event_timestamps(0, 2, 0, rawio._segment_t_start(0, 2), rawio._segment_t_stop(0, 2))
        # Make sure it read 2 events
        self.assertEqual(len(events[0]), 2)

        # Verify the second event of the second segment is a Neuralynx type event, with correct time
        self.assertEqual(events[2][1][:4], "NlxP")
        self.assertEqual(events[0][1], 161.607036)

        rawio.close()

        # Test on second test dataset, test.medd, with preserving original timestamps.
        # Timestamps here are in UTC (seconds since midnight, 1 Jan 1970)
        filename = self.get_local_path("med/test.medd")

        rawio = MedRawIO(filename, password="L2_password", keep_original_times=True)
        rawio.parse_header()

        # Segment 0
        self.assertEqual(rawio._segment_t_start(0, 0), 1678111774.012236)
        self.assertEqual(rawio._segment_t_stop(0, 0), 1678111813.902036)
        # Segment 1
        self.assertEqual(rawio._segment_t_start(0, 1), 1678111824.822062)
        self.assertEqual(rawio._segment_t_stop(0, 1), 1678111861.349882)
        # Segment 2
        self.assertEqual(rawio._segment_t_start(0, 2), 1678111871.254293)
        self.assertEqual(rawio._segment_t_stop(0, 2), 1678111954.028938)

        # Verify that there are 5 events in the dataset.
        # They are 3 "Note" type events, and 2 "NlxP", or neuralynx, type events.
        # The first segment has one event, and the second and third segments
        # each have 2 events.
        self.assertEqual(rawio.event_count(0, 0, 0), 1)
        self.assertEqual(rawio.event_count(0, 1, 0), 2)
        self.assertEqual(rawio.event_count(0, 2, 0), 2)

        # Get array of all events in first segment of data
        events = rawio._get_event_timestamps(0, 0, 0, rawio._segment_t_start(0, 0), rawio._segment_t_stop(0, 0))
        # Make sure it read 1 event
        self.assertEqual(len(events[0]), 1)

        # Get array of all events in second segment of data
        events = rawio._get_event_timestamps(0, 1, 0, rawio._segment_t_start(0, 1), rawio._segment_t_stop(0, 1))
        # Make sure it read 2 events
        self.assertEqual(len(events[0]), 2)

        # Verify the first event of the second segment is a Neuralynx type event, with correct time
        self.assertEqual(events[2][0][:4], "NlxP")
        self.assertEqual(events[0][0], 1678111825.715745)

        # Get array of all events in third segment of data
        events = rawio._get_event_timestamps(0, 2, 0, rawio._segment_t_start(0, 2), rawio._segment_t_stop(0, 2))
        # Make sure it read 2 events
        self.assertEqual(len(events[0]), 2)

        # Verify the second event of the second segment is a Neuralynx type event, with correct time
        self.assertEqual(events[2][1][:4], "NlxP")
        self.assertEqual(events[0][1], 1678111935.619272)

        rawio.close()


if __name__ == "__main__":
    unittest.main()



from neo.io.basefromrawio import BaseFromRaw
from neo.rawio.phyrawio import PhyRawIO


class PhyIO(PhyRawIO, BaseFromRaw):
    name = "Phy IO"
    description = "Phy IO"
    mode = "dir"

    def __init__(self, dirname, load_amplitudes=False, load_pcs=False):
        PhyRawIO.__init__(self, dirname=dirname, load_amplitudes=load_amplitudes, load_pcs=load_pcs)
        BaseFromRaw.__init__(self, dirname)



"""
This module implements :class:`BaseSignal`, an array of signals.
This is a parent class from which all signal objects inherit:
    :class:`AnalogSignal` and :class:`IrregularlySampledSignal`

:class:`BaseSignal` inherits from :class:`quantities.Quantity`, which
inherits from :class:`numpy.array`.
Inheritance from :class:`numpy.array` is explained here:
http://docs.scipy.org/doc/numpy/user/basics.subclassing.html

In brief:
* Constructor :meth:`__new__` for :class:`BaseSignal` doesn't exist.
Only child objects :class:`AnalogSignal` and :class:`IrregularlySampledSignal`
can be created.
"""

import logging
from copy import deepcopy

import numpy as np
import quantities as pq

from neo.core.baseneo import MergeError, merge_annotations
from neo.core.dataobject import DataObject, ArrayDict

logger = logging.getLogger("Neo")


class BaseSignal(DataObject):
    """
    This is the base class from which all signal objects inherit:
    :class:`AnalogSignal` and :class:`IrregularlySampledSignal`.

    This class contains all common methods of both child classes.
    It uses the following child class attributes:

        :_necessary_attrs: a list of the attributes that the class must have.

        :_recommended_attrs: a list of the attributes that the class may
        optionally have.
    """

    def _array_finalize_spec(self, obj):
        """
        Called by :meth:`__array_finalize__`, used to customize behaviour of sub-classes.
        """
        return obj

    def __array_finalize__(self, obj):
        """
        This is called every time a new signal is created.

        It is the appropriate place to set default values for attributes
        for a signal constructed by slicing or viewing.

        User-specified values are only relevant for construction from
        constructor, and these are set in __new__ in the child object.
        Then they are just copied over here. Default values for the
        specific attributes for subclasses (:class:`AnalogSignal`
        and :class:`IrregularlySampledSignal`) are set in
        :meth:`_array_finalize_spec`
        """
        super().__array_finalize__(obj)
        self._array_finalize_spec(obj)

        # The additional arguments
        self.annotations = getattr(obj, "annotations", {})
        # Add empty array annotations, because they cannot always be copied,
        # but do not overwrite existing ones from slicing etc.
        # This ensures the attribute exists
        if not hasattr(self, "array_annotations"):
            self.array_annotations = ArrayDict(self._get_arr_ann_length())

        # Globally recommended attributes
        self.name = getattr(obj, "name", None)
        self.file_origin = getattr(obj, "file_origin", None)
        self.description = getattr(obj, "description", None)

        # Parent objects
        self.segment = getattr(obj, "segment", None)

    @classmethod
    def _rescale(self, signal, units=None):
        """
        Check that units are present, and rescale the signal if necessary.
        This is called whenever a new signal is
        created from the constructor. See :meth:`__new__' in
        :class:`AnalogSignal` and :class:`IrregularlySampledSignal`
        """
        if units is None:
            if not hasattr(signal, "units"):
                raise ValueError("Units must be specified")
        elif isinstance(signal, pq.Quantity):
            # This test always returns True, i.e. rescaling is always executed if one of the units
            # is a pq.CompoundUnit. This is fine because rescaling is correct anyway.
            if pq.quantity.validate_dimensionality(units) != signal.dimensionality:
                signal = signal.rescale(units)
        return signal

    def __getslice__(self, i, j):
        """
        Get a slice from :attr:`i` to :attr:`j`.attr[0]

        Doesn't get called in Python 3, :meth:`__getitem__` is called instead
        """
        return self.__getitem__(slice(i, j))

    def __ne__(self, other):
        """
        Non-equality test (!=)
        """
        return not self.__eq__(other)

    def _apply_operator(self, other, op, *args):
        """
        Handle copying metadata to the new signal
        after a mathematical operation.
        """
        self._check_consistency(other)
        f = getattr(super(), op)
        new_signal = f(other, *args)
        new_signal._copy_data_complement(self)
        # _copy_data_complement can't always copy array annotations,
        # so this needs to be done locally
        new_signal.array_annotations = deepcopy(self.array_annotations)
        return new_signal

    def _get_required_attributes(self, signal, units):
        """
        Return a list of the required attributes for a signal as a dictionary
        """
        required_attributes = {}
        for attr in self._necessary_attrs:
            if attr[0] == "signal":
                required_attributes["signal"] = signal
            elif attr[0] == "image_data":
                required_attributes["image_data"] = signal
            elif attr[0] == "t_start":
                required_attributes["t_start"] = getattr(self, "t_start", 0.0 * pq.ms)
            else:
                required_attributes[str(attr[0])] = getattr(self, attr[0], None)
        required_attributes["units"] = units
        return required_attributes

    def duplicate_with_new_data(self, signal, units=None):
        """
        Create a new signal with the same metadata but different data.
        Required attributes of the signal are used.
        Note: Array annotations can not be copied here because length of data can change
        """
        if units is None:
            units = self.units
        # else:
        #     units = pq.quantity.validate_dimensionality(units)

        # signal is the new signal
        required_attributes = self._get_required_attributes(signal, units)
        new = self.__class__(**required_attributes)
        new._copy_data_complement(self)
        new.annotations.update(self.annotations)
        # Note: Array annotations are not copied here, because it is not ensured
        # that the same number of signals is used and they would possibly make no sense
        # when combined with another signal
        return new

    def _copy_data_complement(self, other):
        """
        Copy the metadata from another signal.
        Required and recommended attributes of the signal are used.
        Note: Array annotations can not be copied here because length of data can change
        """
        all_attr = {self._recommended_attrs, self._necessary_attrs}
        for sub_at in all_attr:
            for attr in sub_at:
                if attr[0] == "t_start":
                    setattr(self, attr[0], deepcopy(getattr(other, attr[0], 0.0 * pq.ms)))
                elif attr[0] != "signal":
                    setattr(self, attr[0], deepcopy(getattr(other, attr[0], None)))
        setattr(self, "annotations", deepcopy(getattr(other, "annotations", None)))

        # Note: Array annotations cannot be copied because length of data can be changed  # here
        #  which would cause inconsistencies

    def __rsub__(self, other, *args):
        """
        Backwards subtraction (other-self)
        """
        return self.__mul__(-1, *args) + other

    def __add__(self, other, *args):
        """
        Addition (+)
        """
        return self._apply_operator(other, "__add__", *args)

    def __sub__(self, other, *args):
        """
        Subtraction (-)
        """
        return self._apply_operator(other, "__sub__", *args)

    def __mul__(self, other, *args):
        """
        Multiplication (*)
        """
        return self._apply_operator(other, "__mul__", *args)

    def __truediv__(self, other, *args):
        """
        Float division (/)
        """
        return self._apply_operator(other, "__truediv__", *args)

    def __div__(self, other, *args):
        """
        Integer division (//)
        """
        return self._apply_operator(other, "__div__", *args)

    __radd__ = __add__
    __rmul__ = __sub__

    def merge(self, other):
        """
        Merge another signal into this one.

        The signal objects are concatenated horizontally
        (column-wise, :func:`np.hstack`).

        If the attributes of the two signal are not
        compatible, an Exception is raised.

        Required attributes of the signal are used.
        """

        for attr in self._necessary_attrs:
            if "signal" != attr[0]:
                if getattr(self, attr[0], None) != getattr(other, attr[0], None):
                    raise MergeError(f"Cannot merge these two signals as the {attr[0]} differ.")

        if self.segment != other.segment:
            raise MergeError("Cannot merge these two signals as they belong to different segments.")
        if hasattr(self, "lazy_shape"):
            if hasattr(other, "lazy_shape"):
                if self.lazy_shape[0] != other.lazy_shape[0]:
                    raise MergeError("Cannot merge signals of different length.")
                merged_lazy_shape = (self.lazy_shape[0], self.lazy_shape[1] + other.lazy_shape[1])
            else:
                raise MergeError("Cannot merge a lazy object with a real object.")
        if other.units != self.units:
            other = other.rescale(self.units)
        stack = np.hstack((self.magnitude, other.magnitude))
        kwargs = {}
        for name in ("name", "description", "file_origin"):
            attr_self = getattr(self, name)
            attr_other = getattr(other, name)
            if attr_self == attr_other:
                kwargs[name] = attr_self
            else:
                kwargs[name] = f"merge({attr_self}, {attr_other})"
        merged_annotations = merge_annotations(self.annotations, other.annotations)
        kwargs.update(merged_annotations)

        kwargs["array_annotations"] = self._merge_array_annotations(other)

        signal = self.__class__(
            stack,
            units=self.units,
            dtype=self.dtype,
            copy=False,
            t_start=self.t_start,
            sampling_rate=self.sampling_rate,
            **kwargs,
        )
        signal.segment = self.segment

        if hasattr(self, "lazy_shape"):
            signal.lazy_shape = merged_lazy_shape

        return signal

    def time_slice(self, t_start, t_stop):
        """
        Creates a new AnalogSignal corresponding to the time slice of the
        original Signal between times t_start, t_stop.
        """
        NotImplementedError("Needs to be implemented for subclasses.")

    def concatenate(self, *signals):
        """
        Concatenate multiple signals across time.

        The signal objects are concatenated vertically
        (row-wise, :func:`np.vstack`). Concatenation can be
        used to combine signals across segments.
        Note: Only (array) annotations common to
        both signals are attached to the concatenated signal.

        If the attributes of the signals are not
        compatible, an Exception is raised.

        Parameters
        ----------
        signals : multiple neo.BaseSignal objects
            The objects that is concatenated with this one.

        Returns
        -------
        signal : neo.BaseSignal
            Signal containing all non-overlapping samples of
            the source signals.

        Raises
        ------
        MergeError
            If `other` object has incompatible attributes.
        """

        NotImplementedError("Patching need to be implemented in subclasses")
